from django.core.files.base import ContentFile
from django.core.files.storage import Storage

storage = dict()


class CustomStorage(Storage):
    def exists(self, key):
        return key in storage

    def _save(self, name, content):
        if self.exists(name):
            raise AssertionError('File already exists, not overwriting it')
        storage[name] = content.file.read()
        return name

    def _open(self, name, mode='rb'):
        return ContentFile(storage[name].encode('utf-8'))



import uuid

from django.core.files.storage import default_storage
from django.db import connection
from django.test import TestCase

from unittest import skip

from storage_text_field.tests.models import (
    CustomStorageDocument,
    CustomStorageObjectDocument,
    Document,
    FromDBHookDocument,
    PreSaveHookDocument,
    UseCacheDocument,
    from_db_hook,
    pre_save_hook,
)
from storage_text_field.tests import storages


class BaseTestCase(object):
    @property
    def html(self):
        return '<p>{}</p>'.format(uuid.uuid4())

    def pre_save_hook(self, value):
        return value

    def from_db_hook(self, value):
        return value

    def get_filepath_from_db(self, obj):
        cursor = connection.cursor()

        # You are not supposed to use string formatting when doing SQL queries,
        # but I hate the way we are supposed to do this. Since this is a test,
        # there is no chance of a SQL injection.
        query = 'SELECT html from {db_table} WHERE id = {pk}'
        cursor.execute(query.format(
            db_table=obj._meta.db_table,
            pk=obj.pk,
        ))
        file_path = cursor.fetchone()[0]
        return file_path

    def format_expected_value(self, value):
        return self.from_db_hook(self.pre_save_hook(value))

    def test_smoke_test(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        document.refresh_from_db()
        self.assertEqual(document.html, self.format_expected_value(html))

    def test_field_can_be_blank(self):
        document = self.Model.objects.create()
        document.refresh_from_db()
        self.assertEqual(document.html, self.format_expected_value(''))

    def test_only_save_on_storage_once(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        document.refresh_from_db()
        original_file_path = self.get_filepath_from_db(document)
        document.save()
        self.assertEqual(
            original_file_path,
            self.get_filepath_from_db(document),
        )

    def test_changing_the_content_changes_the_filepath(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        document.refresh_from_db()
        original_file_path = self.get_filepath_from_db(document)

        # self.html creates new custom HTML.
        document.html = self.html
        document.save()
        self.assertNotEqual(
            original_file_path,
            self.get_filepath_from_db(document),
        )


class SmokeTestCase(BaseTestCase, TestCase):
    Model = Document

    def test_db_is_saving_file_path_to_content(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )

        html_file_path = self.get_filepath_from_db(document)
        with default_storage.open(html_file_path) as f:
            file_contents = f.read().decode('utf-8')
        self.assertEqual(file_contents, html)


class CustomStorageTestCase(BaseTestCase, TestCase):
    Model = CustomStorageDocument

    def setUp(self):
        super(CustomStorageTestCase, self).setUp()
        # Clear the custom storage.
        storages.storage = dict()

    def test_custom_storage_is_used(self):
        html = self.html
        self.Model.objects.create(
            html=html,
        )
        self.assertEqual(len(storages.storage), 1)
        self.assertEqual(
            list(storages.storage.values())[0],
            html,
        )


class CustomStorageObjectTestCase(CustomStorageTestCase):
    Model = CustomStorageObjectDocument


class PreSaveHookTestCase(BaseTestCase, TestCase):
    Model = PreSaveHookDocument

    def pre_save_hook(self, value):
        return pre_save_hook(value)

    @skip('This test does not work with only the pre save hook')
    def test_only_save_on_storage_once(self):
        pass

    def test_html_is_updated_because_of_pre_save_hook(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        document.refresh_from_db()
        self.assertEqual(document.html, self.pre_save_hook(html))


class FromDBHookTestCase(BaseTestCase, TestCase):
    Model = FromDBHookDocument

    def from_db_hook(self, value):
        return from_db_hook(value)

    @skip('This test does not work with only the pre save hook')
    def test_only_save_on_storage_once(self):
        pass

    def test_html_is_updated_because_of_from_db_hook(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        document.refresh_from_db()
        self.assertEqual(document.html, self.from_db_hook(html))


class UseCacheTestCase(BaseTestCase, TestCase):
    Model = UseCacheDocument

    def test_cache_is_used(self):
        html = self.html
        document = self.Model.objects.create(
            html=html,
        )
        # Pull html from the storage to prime the cache.
        document = self.Model.objects.get(pk=document.pk)
        self.assertEqual(document.html, html)

        # Remove the content from the storage.
        storage = document._meta.get_field('html').storage
        original_file_path = self.get_filepath_from_db(document)
        storage.delete(original_file_path)

        # This would fail if the cache was not being used.
        document = self.Model.objects.get(pk=document.pk)
        self.assertEqual(document.html, html)



import hashlib
import os
import six

from django.conf import settings
from django.core.cache import caches
from django.core.files.base import ContentFile
from django.core.files.storage import (
    Storage,
    default_storage,
    get_storage_class,
)
from django.db import models
from django.utils.functional import lazy


def default_pre_save_hook(value):
    return value


def default_from_db_hook(value):
    return value


def default_file_path_hook(value):
    return value


class StorageTextField(models.CharField):
    def __init__(
            self,
            storage=None,
            pre_save_hook=None,
            from_db_hook=None,
            file_path_hook=None,
            use_cache=False,
            *args,
            **kwargs
    ):
        if storage is None:
            storage = default_storage
        if not isinstance(storage, Storage):
            storage = get_storage_class(storage)()
        if not pre_save_hook:
            pre_save_hook = default_pre_save_hook
        if not from_db_hook:
            from_db_hook = default_from_db_hook
        if not file_path_hook:
            file_path_hook = default_file_path_hook
        self.storage = storage
        self.pre_save_hook = pre_save_hook
        self.from_db_hook = from_db_hook
        self.file_path_hook = file_path_hook
        self.use_cache = use_cache
        if self.use_cache:
            self.cache_config = getattr(
                settings,
                'STORAGE_TEXT_FIELD_CACHE_CONFIG',
                'default',
            )
            self.cache_duration = getattr(
                settings,
                'STORAGE_TEXT_FIELD_CACHE_DURATION',
                60 * 60 * 24,  # One Day
            )
            self.cache = caches[self.cache_config]
        kwargs['max_length'] = 200
        super(StorageTextField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(StorageTextField, self).deconstruct()
        del kwargs['max_length']
        kwargs['pre_save_hook'] = self.pre_save_hook
        kwargs['from_db_hook'] = self.from_db_hook
        kwargs['file_path_hook'] = self.file_path_hook
        kwargs['storage'] = self.storage
        kwargs['use_cache'] = self.use_cache
        return name, path, args, kwargs

    def get_file_path(self, value):
        str_self = str(self).encode('utf-8')
        digest = hashlib.sha224(
            str_self + six.text_type(value).encode('utf-8')
        ).hexdigest()
        return self.file_path_hook(os.path.join(
            digest.encode('utf-8'),
            str_self,
        ))

    def get_prep_value(self, value):
        value = self.pre_save_hook(value)
        file_path = self.get_file_path(value).decode('utf-8')
        if not self.storage.exists(file_path):
            file_path = self.storage.save(
                file_path,
                ContentFile(value),
            )
        return super(StorageTextField, self).get_prep_value(file_path)

    def from_db_value(self, value, expression, connection, context):
        def get_from_storage():
            return self.storage.open(value).read()

        def result():
            if self.use_cache:
                content = self.cache.get(value)
                if content is None:
                    content = get_from_storage()
            else:
                content = get_from_storage()
            content = self.from_db_hook(content)
            try:
                return_value = six.text_type(content.decode('utf-8'))
            except (UnicodeEncodeError, AttributeError):
                return_value = six.text_type(content)
            if self.use_cache:
                self.cache.set(value, return_value, self.cache_duration)
            return return_value

        return lazy(result, six.text_type)()



from django.db import models

from storage_text_field.fields import StorageTextField
from storage_text_field.tests.storages import CustomStorage


class Document(models.Model):
    html = StorageTextField()


class CustomStorageDocument(models.Model):
    html = StorageTextField(
        storage='storage_text_field.tests.storages.CustomStorage',
    )


class CustomStorageObjectDocument(models.Model):
    html = StorageTextField(
        storage=CustomStorage(),
    )


def pre_save_hook(value):
    return value + '<strong>FOOBAR</strong>'


class PreSaveHookDocument(models.Model):
    html = StorageTextField(
        pre_save_hook=pre_save_hook,
    )


def from_db_hook(value):
    # Strip off the p tag.
    return value[3:-4]


class FromDBHookDocument(models.Model):
    html = StorageTextField(
        from_db_hook=from_db_hook,
    )


class UseCacheDocument(models.Model):
    html = StorageTextField(use_cache=True)



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name='django-storage-text-field',
    version="1.5.0",
    package_dir={'storage_text_field': 'storage_text_field'},
    packages=['storage_text_field', 'storage_text_field.tests'],
    description='Custom Django field that saves content to storages and a reference to the stored file in the database.',  # noqa
    author='Jason Ward',
    author_email='jason.ward@policystat.com',
    license='MIT License',
    url='http://github.com/PolicyStat/django-storage-text-field/',
    long_description=long_description,
    platforms=["any"],
    install_requiers=[
        'six>=1.10.0,<2',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Environment :: Web Environment',
    ],
)



import os
import uuid


SECRET_KEY = str(uuid.uuid4())

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}


INSTALLED_APPS = [
    'storage_text_field.tests',
]

MEDIA_ROOT = os.path.join(
    os.path.dirname(__file__),
    'media_root',
)
